#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import QUB
import qubtree
from math import *
import random
import __main__ # to evaluate custom shape functions with user's REP defs


# Shape.properties                   :qubtree.Node
#      .set(propname, valuestr)
#      .get(propname)                :valuestr
#      .duration()                   :seconds
#      .generate(first, last, t_start, sampling, rep_index)  :qubtree.Node with double data
#         (first, last are point indices; t_start, sampling in seconds)
#      .generateInto(first, last, t_start, sampling, rep_index, node, i_start)


# the chicken and the egg:
#   we would like to add a new shape like >>> QUB.AddProtocolShape( ShapeClass )
#   but typically the PythonScripts are read in before any other viewport is created,
#   and a function like AddProtocolShape is set up when the Protocol viewport is created.
#
#   so, AddShape will call AddProtocolShape if it exists, otherwise it will queue them for delphi to pick up at its leisure.

ShapeQueue = []

def AddShape(ShapeClass):
    try:
        QUB.AddProtocolShape(ShapeClass)
    except:
        ShapeQueue.append(ShapeClass)


def SampleCount(t_start, sampling, duration):
    small_enough = 0.49 * sampling
    t_start = t_start % sampling
    if t_start < small_enough:
        result = 1
    else:
        result = 0
    end = (t_start + duration) / sampling
    iend = int(end)
    result = result + iend
    if abs(end - iend) < 0.5:
        result = result - 1
    if result < 0:
        result = 0
    return result

# for Base, incr, factor this returns (base+incr*index) * factor ^ index, 
# not   (((Base+Incr)*factor+Incr)*factor
def IncrAndScale(base, incr, factor, rep_index):
    return (base + incr * rep_index) * pow(factor, rep_index)


#---------- Base Shape Class
class Shape:
    name = '(missing)'
    
    def __init__(self, properties=None, defaults=[('Duration', 1.0)]):
        if not properties:
            properties = qubtree.Node('Shape')
        if properties.name != 'Shape':
            properties.name = 'Shape'
        if str(properties.data) != self.name:
            properties.data = self.name
        for prop in defaults:
            if properties.find(prop[0]).isNull:
                properties[prop[0]].data = prop[1]
        self.properties = properties

        self.setup()        

    def renameProperties(self, properties, renames):
        if properties:
            for oldName in renames.keys():
                properties.find(oldName).name = renames[oldName]
                # if not found (null) then setting name does nothing

    def get(self, prop):
        rtn = '0.0'
        try:
            rtn = str(self.properties[prop].data)
        except:
            pass
        return rtn

    def set(self, prop, valstr):
        propNode = self.properties.find(prop)
        if (propNode.data.type == qubtree.QTR_TYPE_STRING) or (propNode.data.type == qubtree.QTR_TYPE_EMPTY):
            propNode.data = valstr
        elif propNode.data.count == 1:
            try:
                if (propNode.data.type == qubtree.QTR_TYPE_FLOAT) or (propNode.data.type == qubtree.QTR_TYPE_DOUBLE):
                    propNode.data[0] = float(eval(valstr))
                else:
                    propNode.data[0] = int(eval(valstr))
            except:
                pass

        self.setup()            

    def setup(self):        
        # pre-evaluate the time-crucial stuff
        pass

    def duration(self, rep_index):
        return self.properties['Duration'].data[0]

    def count(self, t_start, sampling, rep_index):
        return SampleCount(t_start, sampling, self.duration(rep_index))

    def evaluate(self, t, rep_index):
        return 0.0
    
    def generate(self, first, last, t_start, sampling, rep_index):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1

        node = qubtree.Node()
        node.data.setup(qubtree.QTR_TYPE_DOUBLE, count, 1)
        
        self.generateInto(first, last, t_start, sampling, rep_index, node, 0)
        return node

    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['sampling'].data = sampling
        params['t_off'].data = t_off
        params['first'].data = first
        params['remain'].data = remain
        params['rep_index'].data = rep_index
        params['i_start'].data = i_start

        QUB.ProtocolShape_GenerateInto(self.evaluate, node, params)
        # compiled loop of the following:
        #data = node.data
        #for i in xrange(0, remain):
        #    data[i_start + i] = self.evaluate(t_off + (first + i) * sampling, rep_index)

#---------- Custom definded function
class CustomShape(Shape):
    name = 'Custom'
    
    def __init__(self, properties=None):
        self.renameProperties(properties, {'f(x)' : 'f(x,i)'})
        defaults = [('f(x,i)', 'x * x'),
                    ('x from', 0.0),
                    ('x to', 1.0),
                    ('Duration', 1.0)]
        Shape.__init__(self, properties, defaults)

    def setup(self):
        try:
            self.xfrom = self.properties['x from'].data[0]
            self.xto = self.properties['x to'].data[0]
            self.dur = self.properties['Duration'].data[0]
            self.xfact = (self.xto - self.xfrom) / self.dur
            self.xfunc = eval('lambda x, i: ' + str(self.properties['f(x,i)'].data), __main__.__dict__)
            self.tfunc = lambda t, i: self.xfunc( t * self.xfact + self.xfrom, i )
        except:
            self.tfunc = lambda t, i: 0

    def duration(self, rep_index):
        return self.dur

    def evaluate(self, t, rep_index):
        return self.tfunc(t, rep_index)

#---------- step function 
class Step(Shape):
    name = 'Step'
    
    def __init__(self, properties=None):
        self.currRep = 0
        
        self.renameProperties(properties, {'predur' : 'preDuration',
                                           'prelvl' : 'preAmp',
                                           'dur'    : 'Duration',
                                           'lvl'    : 'Amp',
                                           'delta_lvl' : 'Amp incr',
                                           'deltaAmp'  : 'Amp incr',
                                           'postdur' : 'postDuration',
                                           'postlvl' : 'postAmp'})
        defaults = [('Duration', 0.5),
                    ('Duration incr', 0.0),
                    ('Duration factor', 1.0),
                    ('Amp', 1.0),
                    ('Amp incr', 0.0),
                    ('Amp factor', 1.0),
                    ('preDuration', 0.25),
                    ('preDuration incr', 0.0),
                    ('preDuration factor', 1.0),
                    ('preAmp', 0.0),
                    ('preAmp incr', 0.0),
                    ('preAmp factor', 1.0),
                    ('postDuration', 0.25),
                    ('postDuration incr', 0.0),
                    ('postDuration factor', 1.0),
                    ('postAmp', 0.0),
                    ('postAmp incr', 0.0),
                    ('postAmp factor', 1.0)]
        
        Shape.__init__(self, properties, defaults)

    def setup(self):
        # pre-evaluate the time-crucial stuff
        preDur = IncrAndScale( self.properties['preDuration'].data[0],
                               self.properties['preDuration incr'].data[0],
                               self.properties['preDuration factor'].data[0],
                               self.currRep )
        preAmp = IncrAndScale( self.properties['preAmp'].data[0],
                               self.properties['preAmp incr'].data[0],
                               self.properties['preAmp factor'].data[0],
                               self.currRep )
        pre2Dur = preDur + IncrAndScale( self.properties['Duration'].data[0],
                                         self.properties['Duration incr'].data[0],
                                         self.properties['Duration factor'].data[0],
                                         self.currRep )
        amp = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            self.currRep )
        postAmp = IncrAndScale( self.properties['postAmp'].data[0],
                                self.properties['postAmp incr'].data[0],
                                self.properties['postAmp factor'].data[0],
                                self.currRep )

        def tfunc(t):
            if t < preDur:
                return preAmp
            elif t < pre2Dur:
                return amp
            else:
                return postAmp
        self.tfunc = tfunc        

    def duration(self, rep_index):
        return (  IncrAndScale( self.properties['preDuration'].data[0],
                                self.properties['preDuration incr'].data[0],
                                self.properties['preDuration factor'].data[0],
                                rep_index )
                + IncrAndScale( self.properties['Duration'].data[0],
                                self.properties['Duration incr'].data[0],
                                self.properties['Duration factor'].data[0],
                                rep_index )
                + IncrAndScale( self.properties['postDuration'].data[0],
                                self.properties['postDuration incr'].data[0],
                                self.properties['postDuration factor'].data[0],
                                rep_index ) )

    def evaluate(self, t, rep_index):
        if rep_index != self.currRep:
            self.currRep = rep_index
            self.setup()
        return self.tfunc(t)

    # generateInto: calls compiled loop. makes setup and evaluate irrelevant.
    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['preDur'].data = IncrAndScale( self.properties['preDuration'].data[0],
                            self.properties['preDuration incr'].data[0],
                            self.properties['preDuration factor'].data[0],
                            rep_index )
        params['preAmp'].data = IncrAndScale( self.properties['preAmp'].data[0],
                            self.properties['preAmp incr'].data[0],
                            self.properties['preAmp factor'].data[0],
                            rep_index )
        params['pre2Dur'].data = params['preDur'].data[0] + IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            rep_index )
        params['amp'].data = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            rep_index )
        params['postAmp'].data = IncrAndScale( self.properties['postAmp'].data[0],
                            self.properties['postAmp incr'].data[0],
                            self.properties['postAmp factor'].data[0],
                            rep_index )
        params['sampling'].data = sampling
        params['t_off'].data = t_off
        params['first'].data = first
        params['remain'].data = remain
        params['i_start'].data = i_start

        QUB.ProtocolShape_Step_GenerateInto(node, params)

#---------- Sine( amp, freq, offset)
class Sine(Shape):
    name = 'Sine'
    
    def __init__(self, properties=None):
        self.currRep = 0        

        self.renameProperties(properties, {'dur'            : 'Duration',
                                           'ampl'           : 'Amp',
                                           'delta_ampl'     : 'Amp incr',
                                           'deltaAmp'       : 'Amp incr',
                                           'freq'           : 'Freq',
                                           'delta_freq'     : 'Freq incr',
                                           'deltaFreq'      : 'Freq incr',
                                           't_offset'       : 'Phase',
                                           'delta_t_offset' : 'Phase incr',
                                           'deltaPhase'     : 'Phase incr'})
        defaults = [('Duration', 1.0),
                    ('Duration incr', 0.0),
                    ('Duration factor', 1.0),
                    ('Amp', 1.0),
                    ('Amp incr', 0.0),
                    ('Amp factor', 1.0),
                    ('Freq', 60.0),
                    ('Freq incr', 0.0),
                    ('Freq factor', 1.0),
                    ('Phase', 0.0),
                    ('Phase incr', 0.0),
                    ('Phase factor', 1.0),
                    ('Offset', 0.0),
                    ('Offset incr', 0.0),
                    ('Offset factor', 1.0)]
        
        Shape.__init__(self, properties, defaults)

    def setup(self):        
        # pre-evaluate the time-crucial stuff
        dur = IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            self.currRep )
        amp = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            self.currRep )
        freq2pi = 2 * pi * IncrAndScale( self.properties['Freq'].data[0],
                             self.properties['Freq incr'].data[0],
                             self.properties['Freq factor'].data[0],
                             self.currRep )
        phase = IncrAndScale( self.properties['Phase'].data[0],
                              self.properties['Phase incr'].data[0],
                              self.properties['Phase factor'].data[0],
                              self.currRep )
        off = IncrAndScale( self.properties['Offset'].data[0],
                            self.properties['Offset incr'].data[0],
                            self.properties['Offset factor'].data[0],
                            self.currRep )
        pi2 = 2 * pi
        self.tfunc = lambda t: off + amp * sin(t * freq2pi + phase)

    def duration(self, rep_index):
        return IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            rep_index )

    def evaluate(self, t, rep_index):
        if rep_index != self.currRep:
            self.currRep = rep_index
            self.setup()
        return self.tfunc(t)

    # generateInto: calls compiled loop. makes setup and evaluate irrelevant.
    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['amp'].data = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            rep_index )
        params['freq'].data = IncrAndScale( self.properties['Freq'].data[0],
                             self.properties['Freq incr'].data[0],
                             self.properties['Freq factor'].data[0],
                             rep_index )
        params['phase'].data = IncrAndScale( self.properties['Phase'].data[0],
                              self.properties['Phase incr'].data[0],
                              self.properties['Phase factor'].data[0],
                              rep_index )
        params['off'].data = IncrAndScale( self.properties['Offset'].data[0],
                            self.properties['Offset incr'].data[0],
                            self.properties['Offset factor'].data[0],
                            rep_index )
        params['sampling'].data = sampling
        params['t_off'].data = t_off
        params['first'].data = first
        params['remain'].data = remain
        params['i_start'].data = i_start

        QUB.ProtocolShape_Sine_GenerateInto(node, params)
            

#----- Ramp - Straight line rise or fall 
class Ramp(Shape):
    name = 'Ramp'
    
    def __init__(self, properties=None):
        self.currRep = 0        

        self.renameProperties(properties, {'dur'            : 'Duration',
                                           'from_lvl'       : 'FromAmp',
                                           'to_lvl'         : 'ToAmp',
                                           'delta_to'       : 'ToAmp incr',
                                           'deltaToAmp'     : 'ToAmp incr'})
        defaults = [('Duration', 1.0),
                    ('Duration incr', 0.0),
                    ('Duration factor', 1.0),
                    ('FromAmp', 0.0),
                    ('FromAmp incr', 0.0),
                    ('FromAmp factor', 1.0),
                    ('ToAmp', 1.0),
                    ('ToAmp incr', 0.0),
                    ('ToAmp factor', 1.0)]
        
        Shape.__init__(self, properties, defaults)

    def setup(self):        
        # pre-evaluate the time-crucial stuff
        dur = IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            self.currRep )
        frm = IncrAndScale( self.properties['FromAmp'].data[0],
                            self.properties['FromAmp incr'].data[0],
                            self.properties['FromAmp factor'].data[0],
                            self.currRep )
        to = IncrAndScale( self.properties['ToAmp'].data[0],
                            self.properties['ToAmp incr'].data[0],
                            self.properties['ToAmp factor'].data[0],
                            self.currRep )
        slope = (to - frm) / dur
        self.tfunc = lambda t: frm + t * slope

    def duration(self, rep_index):
        return IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            rep_index )

    def evaluate(self, t, rep_index):
        if rep_index != self.currRep:
            self.currRep = rep_index
            self.setup()
        return self.tfunc(t)

    # generateInto: calls compiled loop. makes setup and evaluate irrelevant.
    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['fromAmp'].data = IncrAndScale( self.properties['FromAmp'].data[0],
                            self.properties['FromAmp incr'].data[0],
                            self.properties['FromAmp factor'].data[0],
                            rep_index )
        params['toAmp'].data = IncrAndScale( self.properties['ToAmp'].data[0],
                             self.properties['ToAmp incr'].data[0],
                             self.properties['ToAmp factor'].data[0],
                             rep_index )
        params['dur'].data = IncrAndScale( self.properties['Duration'].data[0],
                             self.properties['Duration incr'].data[0],
                             self.properties['Duration factor'].data[0],
                             rep_index )
        params['sampling'].data = sampling
        params['t_off'].data = t_off
        params['first'].data = first
        params['remain'].data = remain
        params['i_start'].data = i_start

        QUB.ProtocolShape_Ramp_GenerateInto(node, params)

#---------- Flat line
class Hold(Shape):
    name = 'Hold'
    
    def __init__(self, properties=None):
        self.currRep = 0        

        self.renameProperties(properties, {'dur'            : 'Duration',
                                           'lvl'            : 'Amp',
                                           'deltalvl'       : 'Amp incr',
                                           'deltaAmp'       : 'Amp incr'})
        defaults = [('Duration', 1.0),
                    ('Duration incr', 0.0),
                    ('Duration factor', 1.0),
                    ('Amp', 0.0),
                    ('Amp incr', 0.0),
                    ('Amp factor', 1.0)]
        
        Shape.__init__(self, properties, defaults)

    def setup(self):        
        # pre-evaluate the time-crucial stuff
        amp = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            self.currRep )
        self.tfunc = lambda t: amp

    def duration(self, rep_index):
        return IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            rep_index )

    def evaluate(self, t, rep_index):
        if rep_index != self.currRep:
            self.currRep = rep_index
            self.setup()
        return self.tfunc(t)

    # generateInto: calls compiled loop. makes setup and evaluate irrelevant.
    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['amp'].data = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            rep_index )
        params['first'].data = first
        params['remain'].data = remain
        params['i_start'].data = i_start


        QUB.ProtocolShape_Hold_GenerateInto(node, params)

#---------- Bits - same as Hold with different name 
class Bits(Shape):
    name = 'Bits'
    
    def __init__(self, properties=None):
        self.currRep = 0        

        defaults = [('Duration', 1.0),
                    ('Duration incr', 0.0),
                    ('Duration factor', 1.0),
                    ('Amp', 1.0),
                    ('Amp incr', 0.0),
                    ('Amp factor', 1.0)]
        
        Shape.__init__(self, properties, defaults)

    def setup(self):        
        # pre-evaluate the time-crucial stuff
        amp = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            self.currRep )
        self.tfunc = lambda t: amp

    def duration(self, rep_index):
        return IncrAndScale( self.properties['Duration'].data[0],
                            self.properties['Duration incr'].data[0],
                            self.properties['Duration factor'].data[0],
                            rep_index )

    def evaluate(self, t, rep_index):
        if rep_index != self.currRep:
            self.currRep = rep_index
            self.setup()
        return self.tfunc(t)

    # generateInto: calls compiled loop. makes setup and evaluate irrelevant.
    def generateInto(self, first, last, t_start, sampling, rep_index, node, i_start):
        count = SampleCount(t_start, sampling, self.duration(rep_index))
        first = min(first, count-1)
        last = min(last, count-1)
        if last < 0:
            last = first - 1
        remain = last - first + 1
        t_off = t_start % sampling

        params = qubtree.Node()
        params['amp'].data = IncrAndScale( self.properties['Amp'].data[0],
                            self.properties['Amp incr'].data[0],
                            self.properties['Amp factor'].data[0],
                            rep_index )
        params['first'].data = first
        params['remain'].data = remain
        params['i_start'].data = i_start

        QUB.ProtocolShape_Hold_GenerateInto(node, params)

AddShape( CustomShape )
AddShape( Step )
AddShape( Sine )
AddShape( Ramp )
AddShape( Hold )
AddShape( Bits )

